/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: FileUtils.java 4072 2008-10-05 18:07:54Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.file;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.nio.channels.FileChannel;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Enumeration;
import java.util.Map;
import java.util.Set;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * This class manages operation done many times on files, like copying them.
 * @author Florent Benoit
 */
public final class FileUtils {

    /**
     * Size of the buffer.
     */
    private static final int BUFFER_SIZE = 4096;

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(FileUtils.class);

    /**
     * Date format.
     */
    private static final String DATE_FORMAT = "_yyyy.MM.dd-HH.mm.ss";

    /**
     * Utility class, no public constructor.
     */
    private FileUtils() {

    }

    /**
     * Unpack the source archive in a given directory and returns directory
     * directory created.
     * @param packedJar source JarFile to be unpacked
     * @param dest the destination folder
     * @throws FileUtilsException When unpack fails
     */
    public static void unpack(final JarFile packedJar, final File dest) throws FileUtilsException {

        JarEntry entry = null;

        // get entries of the jar file
        Enumeration<JarEntry> entries = packedJar.entries();
        while (entries.hasMoreElements()) {
            entry = entries.nextElement();

            // File entry
            File entryFile = new File(dest, entry.getName());

            // Create directory
            if (entry.isDirectory()) {
                if (!entryFile.exists()) {
                    // create parent directories (with mkdirs)
                    if (!entryFile.mkdirs()) {
                        String err = "Can not create directory " + entryFile + ", Check the write access.";
                        throw new FileUtilsException(err);
                    }
                }
                continue;
            }

            // If it's a file, we must extract the file
            // Ensure that the directory exists.
            entryFile.getParentFile().mkdirs();

            InputStream is = null;
            // get the input stream
            try {
                is = packedJar.getInputStream(entry);
                // Dump to the file
                dump(is, entryFile);
            } catch (IOException ioe) {
                throw new FileUtilsException("Cannot get inputstream of entry '" + entry + "' of file '" + packedJar + "'.");
            } finally {
                try {
                    is.close();
                } catch (IOException ioe) {
                    logger.error("Cannot close input stream", ioe);
                }
            }

        }
    }

    /**
     * Write the given input stream in the given file.
     * @param in the inputStream to copy.
     * @param entryFile the file where the inputStream must be dumped.
     * @throws FileUtilsException if the dump failed.
     */
    public static void dump(final InputStream in, final File entryFile) throws FileUtilsException {

        try {
            // File output
            FileOutputStream out = new FileOutputStream(entryFile);
            int n = 0;
            try {
                // buffer
                byte[] buffer = new byte[BUFFER_SIZE];
                n = in.read(buffer);
                while (n > 0) {
                    out.write(buffer, 0, n);
                    n = in.read(buffer);
                }
                // Flush
                out.flush();
            } finally {
                out.close();
            }
        } catch (IOException e) {
            String err = "Error while unpacking entry " + entryFile + " : ";
            throw new FileUtilsException(err, e);
        }
    }

    /**
     * Print a text file in a print writer.
     * @param txtFile the text file
     * @param printWriter the print writer
     * @throws FileUtilsException if error occurs
     */
    public static void print(final File txtFile, final PrintWriter printWriter) throws FileUtilsException {
        try {
            BufferedReader bufferedReader = new BufferedReader(new InputStreamReader(new FileInputStream(txtFile)));
            String line = bufferedReader.readLine();
            while (line != null) {
                printWriter.println(line);
                line = bufferedReader.readLine();
            }
            bufferedReader.close();
        } catch (IOException e) {
            throw new FileUtilsException("Error while printing", e);
        }
    }

    /**
     * Print a text file in standard output.
     * @param txtFile the text file
     * @throws FileUtilsException if error occurs
     */
    public static void print(final File txtFile) throws FileUtilsException {
        print(txtFile, new PrintWriter(System.out, true));
    }

    /**
     * Copy a file.
     * @param src source file
     * @param dest dest file
     * @throws FileUtilsException if the copy of the file failed
     */
    public static void copyFile(final String src, final String dest) throws FileUtilsException {
        copyFile(new File(src), new File(dest));
    }

    /**
     * Copy a file.
     * @param src source file
     * @param dest dest file
     * @throws FileUtilsException if the copy of the file failed
     */
    public static void copyFile(final File src, final File dest) throws FileUtilsException {
        File newDest = null;
        if (dest.isDirectory()) {
            logger.debug("Copy a file to a directory, append source filename to directory.");
            newDest = new File(dest, src.getName());
        } else {
            newDest = dest;
        }

        FileInputStream fIn = null;
        FileOutputStream fOut = null;
        FileChannel sourceFC = null;
        FileChannel targetFC = null;
        try {

            // InputStream
            fIn = new FileInputStream(src);
            fOut = new FileOutputStream(newDest);

            // nio channel
            sourceFC = fIn.getChannel();
            targetFC = fOut.getChannel();

            // transfer
            targetFC.transferFrom(sourceFC, 0, sourceFC.size());

            // ensure all has been written
            targetFC.force(true);
        } catch (Exception e) {
            throw new FileUtilsException("Error during copy file : " + src + " -> " + newDest, e);
        } finally {
            try {
                if (sourceFC != null) {
                    sourceFC.close();
                }
                if (targetFC != null) {
                    targetFC.close();
                }

                if (fOut != null) {
                    fOut.flush();
                    fOut.close();
                }
                if (fIn != null) {
                    fIn.close();
                }
            } catch (Exception e) {
                logger.error("Cannot close some i/o which are open.", e);
            }

        }
    }

    /**
     * @param path file/directory to be deleted
     * @return true if deletion was OK
     */
    public static boolean delete(final String path) {
        return delete(new File(path));
    }

    /**
     * @param f file/directory to be deleted
     * @return true if deletion was OK
     */
    public static boolean delete(final File f) {
        if (f.isFile()) {
            return f.delete();
        }
        // else (if directory)
        File[] children = f.listFiles();
        if (children == null) {
            // no children
            return f.delete();
        }

        // There are children
        boolean result = true;
        for (int i = 0; i < children.length; i++) {
            result &= delete(children[i]);
        }
        return result && f.delete();
    }

    /**
     * Copy a directory recursively.
     * @param src source directory
     * @param dest dest directory
     * @throws FileUtilsException if the copy of the directory failed
     */
    public static void copyDirectory(final String src, final String dest) throws FileUtilsException {
        copyDirectory(new File(src), new File(dest));
    }

    /**
     * Copy a directory recursively.
     * @param src source directory
     * @param dest dest directory
     * @throws FileUtilsException if the copy of the directory failed
     */
    public static void copyDirectory(final File src, final File dest) throws FileUtilsException {

        if (!src.isDirectory()) {
            // We don not accept file arguments !
            throw new IllegalArgumentException("Source '" + src + "' must be a directory");
        }

        // create the destination directory if it is inexistant
        if (!dest.exists()) {
            dest.mkdirs();
        }

        // copy the files of the source directory
        File[] childs = src.listFiles();
        if (childs != null) {
            // childs
            for (int i = 0; i < childs.length; i++) {
                File child = childs[i];
                if (child.isFile()) {
                    // file
                    copyFile(child, dest);
                } else {
                    // directory
                    copyDirectory(child, new File(dest, child.getName()));
                }
            }
        }
    }

    /**
     * Change the attribute value in a jar file (and update file if required).
     * @param jarName output jar file
     * @param attributes a list of attributes (name <-> value)
     * @throws FileUtilsException if the manifest is not updated.
     */
    public static void updateAttributesInManifest(final String jarName, final Map<String, String> attributes)
            throws FileUtilsException {

        // First, read the manifest in order to check if attributes needs to be
        // updated or not
        File file = new File(jarName);
        if (!file.exists()) {
            throw new FileUtilsException("The file '" + file + "' is not existing on the filesystem");
        }

        // Create jarFile on it and read manifest
        JarFile jarFile = null;
        Manifest manifest = null;
        try {
            jarFile = new JarFile(file);
            manifest = jarFile.getManifest();
        } catch (IOException e) {
            throw new FileUtilsException("Error while using jar file " + file + "'", e);
        } finally {
            if (jarFile != null) {
                try {
                    jarFile.close();
                } catch (IOException e) {
                    throw new FileUtilsException("Error while closing jar file " + jarFile + "'", e);
                }
            }
        }

        // Check if manifest is up-to-date. If not, update it and change a flag
        boolean manifestUpdated = false;

        if (manifest == null) {
            manifest = new Manifest();
            // Ensure that there is the MANIFEST_VERSION
            Attributes mainAttributes = manifest.getMainAttributes();
            mainAttributes.putValue("Manifest-Version", "1.0");
            manifestUpdated = true;
        }

        // Check given attributes with manifest attributes
        Attributes manifestAttributes = manifest.getMainAttributes();

        Set<Map.Entry<String, String>> entries = attributes.entrySet();
        for (Map.Entry<String, String> entry : entries) {
            String attributeName = entry.getKey();
            String attributeValue = entry.getValue();

            // exist in manifest
            String manifestAttributeValue = manifestAttributes.getValue(attributeName);
            if (attributeValue != null && !attributeValue.equals(manifestAttributeValue)) {
                // Update manifest
                Attributes.Name manifestAttributeName = new Attributes.Name(attributeName);
                manifestAttributes.put(manifestAttributeName, attributeValue);
                manifestUpdated = true;
            }
        }

        // manifest has not been updated, values are up-to-date ? then stop
        // everything
        if (!manifestUpdated) {
            return;
        }

        // Temp file
        File tmpJarFile = null;
        try {
            tmpJarFile = File.createTempFile("tmpGenICFile", ".jar");
            updateJarFile(file, manifest, tmpJarFile);
            FileUtils.copyFile(tmpJarFile, file);
        } catch (IOException e) {
            throw new FileUtilsException("Cannot create temporary file", e);
        } finally {
            tmpJarFile.delete();
        }
    }

    /**
     * For a given file, get the shorter name containing the last modification
     * date.
     * @param file the given file to analyse to get the file name and the last
     *        modification date
     * @return A new archive name which contains the last modification date of
     *         the given deployable
     * @throws FileUtilsException Thrown if the given file cannot be retrieved
     */
    public static String lastModifiedFileName(final File file) throws FileUtilsException {
        if (!file.exists()) {
            throw new FileUtilsException("Cannot get File '" + file + "'.");
        }

        // Get last modification date
        Date lastModificationDate = new Date(file.lastModified());

        // Create a format
        SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);

        return file.getName().replaceAll("\\..{3}$", sdf.format(lastModificationDate) + "$0");
    }

    /**
     * Update the given file with the given manifest and write result in
     * tmpJarFile.
     * @param file the jarfile which have the content
     * @param manifest the manifest to use
     * @param tmpJarFile the resulting file
     * @throws FileUtilsException if jar file is not updated
     */
    private static void updateJarFile(final File file, final Manifest manifest, final File tmpJarFile) throws FileUtilsException {

        // File Stream
        FileOutputStream fileOS = null;
        try {
            fileOS = new FileOutputStream(tmpJarFile);
        } catch (FileNotFoundException e) {
            throw new FileUtilsException("Cannot create file output stream on the file '" + file + "'", e);
        }

        // Creates an ouputStream with the given manifest
        JarOutputStream fileJarOutputStream = null;
        try {
            fileJarOutputStream = new JarOutputStream(fileOS, manifest);
        } catch (IOException e) {
            throw new FileUtilsException("Cannot create Jar OutputStream", e);
        }

        JarFile jarFile = null;
        try {

            int n;
            byte[] buffer = new byte[BUFFER_SIZE];
            jarFile = new JarFile(file);
            Enumeration<JarEntry> jarEntries = jarFile.entries();

            while (jarEntries.hasMoreElements()) {

                // Get Entry
                JarEntry jarEntry = jarEntries.nextElement();

                // Ignore Manifest entry as it is already added by
                // JarOutputStream
                if (JarFile.MANIFEST_NAME.equals(jarEntry.getName())) {
                    continue;
                }

                // Store entry
                fileJarOutputStream.putNextEntry(jarEntry);

                // Put input stream
                InputStream inputStream = jarFile.getInputStream(jarEntry);
                try {

                    n = inputStream.read(buffer);
                    while (n > 0) {
                        fileJarOutputStream.write(buffer, 0, n);
                        n = inputStream.read(buffer);
                    }
                    fileJarOutputStream.closeEntry();
                    // fileJarOutputStream.flush();
                } finally {
                    if (inputStream != null) {
                        inputStream.close();
                    }
                }

            }
        } catch (IOException ioe) {
            throw new FileUtilsException("Cannot update jar file '" + file + "'", ioe);
        } finally {
            // Close streams
            try {
                fileJarOutputStream.flush();
                fileJarOutputStream.close();
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close jar file output stream", e);
            }

            try {
                fileOS.close();
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close file output stream", e);
            }

            try {
                if (jarFile != null) {
                    jarFile.close();
                }
            } catch (IOException e) {
                throw new FileUtilsException("Cannot close jar file '" + jarFile + "'.", e);
            }
        }
    }

}
